#!/bin/sh 
#  \
exec $WISH $0 $@

#################################################
#
# Package to read and to write Excel Biff files
# written by Christoph Bauer 2001
#


# I think MS didn't understand its own record format
# And so you can not write all records completly out of order.
#
# notes about oder to save a biff file
#    * first records are (in this order): BOF FILEPASS INDEXROW
#      FILEPASS and INDEX_ROW are optional (and it is recommend
#      to skip this records)
#    * EXTERNAME follows directly after the associsyed EXTERNSHEET
#      record
#    * do not change the order of the EXTERNSHEET records, you can
#      add more after the last EXTERNSHEET record
#    * all NAME records should appear together. do not change the order
#      of the NAME records
#    * COLUMN_DEFAULT record must appear before any ROW or cell record
#    * STRING record appears after a FORMLA/ARRAY record whose formula
#      evaluates to a string
#    * NOTE record can be splitted. First record: b(len) contains the total
#      length of the notes, the subsequent NOTE record have b(row) == -1 and
#      b(col) == 0.
#    * WINDOW2 record follows the associated WINDOW1 record
#    * SELECTION can be splitted in consecuting records

package require Debug

namespace eval Biff {

    # name of records
    variable version_loaded -1
    variable bigendian 
    if { [binary format f2 {1.6 3.4}] != "\xcd\xcc\xcc\x3f\x9a\x99\x59\x40" } {
        set bigendian 1
        Debug::printMessage information "big-endian machine detected"
    } else {
        set bigendian 0
        Debug::printMessage information "little-endian machine detected"
    }
}

proc Biff::Biff2 {} {
    variable version_loaded
    variable typeNames
    variable typeNums 

    if { $version_loaded == 2 } {
        return 
    }

    if { [info exists typeNames] } {
        #unset typeNames
        #unset typeNums
    }

    array set typeNames {
        9 BOF
        47 FILEPASS
        11 INDEX
        12 CALCCOUNT
        13 CALCMODE
        14 PRECISION
        15 REFMODE
        16 DELTA
        17 ITERATION
        34 1904
        64 BACKUP
        42 PRINT_ROW_HEADER
        43 PRINT_GRIDLINES
        27 HORIZONTAL_PAGE_BREAKS 
        26 VERTICAL_PAGE_BREAKS
        37 DEFAULT_ROW_HEIGHT
        49 FONT
        50 FONT2
        20 HEADER
        21 FOOTER
        38 LEFT_MARGIN
        39 RIGHT_MARGIN
        40 TOP_MARGIN
        41 BOTTOM_MARGIN
        36 COLWIDTH
        22 EXTERNCOUNT
        23 EXTERNSHEET
        35 EXTERNNAME
        30 FORMAT
        24 NAME
        0 DIMENSION
        32 COLUMN_DEFAULT
        8 ROW
        1 BLANK
        2 INTEGER
        3 NUMBER
        5 BOOLERR
        6 FORMULA
        4 LABEL
        33 ARRAY
        60 CONTINUE
        7 STRING
        54 TABLE
        55 TABLE2
        18 PROTECT
        25 WINDOW_PROTECT
        19 PASSWORD
        28 NOTE
        61 WINDOW1
        62 WINDOW2
        65 PANE
        29 SELECTION
        10 EOT
        133 BOUNDSHEET
        224 XF
    }


    # table to translate record names to 
    # the type number

    #variable typeNums
    foreach key [array names typeNames] {
        set typeNums($typeNames($key)) $key
    }
    
    # meanings of b(document_type)
    variable document_typeName
    array set document_typeName {
        16 worksheet
        32 chart
        64 "macro sheet (BIFF4)"
        256 "woorkbook globals (BIFF4)"
        5 "workbooks globals"
        6 "Visual Basic module"
    }
    
    # meanings of b(version_number)
    variable version_numberNam
    array set version_numberName {
        2 "Excel document"
        3 "Multiplan document"
    }

    # flags of b(version_number)
    variable version_numberFlags
    array set version_numberFlags {
        0x0100 bitFMP 
        0xFE00 RESERVED
    }
    
    # meanings of b(calculation_mode)
    variable calculation_modeName
    array set calculation_modeName {
        0 manual
        1 automatic
        -1 automatic_no_tables
    }

    # meanings of b(precision)
    variable precisionName
    array set precisionName {
        1 full_precision
        0 precision_as_displayed
    }

    # meanings of b(reference_mode)
    variable reference_modeName
    array set reference_modeName {
        1 A1
        0 R1C1
    }

    variable pane_numberName
    array set pane_numberName {
        0 "bottom right"
        1 "top right"
        2 "bottom left"
        3 "top left"
    }

    variable font_attrFlags
    array set font_attrFlags { 
        0xFFF0 RESERVED
        0x0008 StikeOut
        0x0004 Underline
        0x0002 Italic
        0x0001 Bold
    }

    variable name_attr1Flags
    array set name_attr1Flags {
        0xF8 RESERVED_F8
        0x04 CalcExpr
        0x02 Proc
        0x01 RESERVED_01
    }

    variable name_attr2Flags
    array set name_attr2Flags {
        0xFC RESERVED
        0x02 Run
        0x01 Func
    }
    
    variable row_heightFlag
    set row_heightFlag {
        0x8000 DefaultHeight
    }

    # possible meaning of b($value) (if b(is_error) == 1)
    variable errorString
    array set errorString {
        0 NULL
        7 DIV/0
        15 VALUE
        23 REF
        29 NAME
        36 NUM
        32 N/A
    }
        
    # parse string given by record name
    variable binRec
    array set binRec {
        BOF {ss b(version_number) b(document_type)}
        FILEPASS {}
        INDEX {issi* b(file_positon_of_first_NAME_record) \
                       b(minRow) b(maxRow) b(file_positon_of_ROW_records) }
        CALCCOUNT {s b(iteration_count)}
        CALCMODE {s b(calculation_mode)}
        PRECISION {s b(precision)}
        REFMODE {s b(reference_mode)}
        DELTA {f b(max_change_for_iteration)}
        ITERATION {s b(iteration_flag)}
        1904 {s b(1904)}
        BACKUP {s b(backup)} 
        PRINT_ROW_HEADER {s b(print_row_header)}
        PRINT_GRIDLINES {s b(print_gridlines)}
        HORIZONTAL_PAGE_BREAKS {ss* b(number_of_horiz_page_breaks)
            b(list_of_rows_after_page_break)}
        VERTICAL_PAGE_BREAKS {ss* b(number_of_vert_page_breaks)
            b(list_of_cols_after_page_break)}
        DEFAULT_ROW_HEIGHT {s b(default_row_height)}
        FONT {ssca* b(font_height) b(font_attr) b(len) b(font_name)}
        FONT2 {}
        HEADER {ca* b(len) b(header_string)}
        FOOTER {ca* b(len) b(footer_string)}
        LEFT_MARGIN {f b(left_margin)}
        RIGHT_MARGIN {f b(right_margin)}
        TOP_MARGIN {f b(top_margin)}
        BOTTOM_MARGIN {f b(bottom_margin)}
        COLWIDTH {ccs b(first_col) b(last_col) b(col_width)}
        EXTERNCOUNT {s b(number_of_extern_references)}
        EXTERNSHEET {cs* b(len) b(extern_sheet)}
        EXTERNNAME {cs* b(len) b(extern_ref_name)}
        FORMAT {cs* b(len) b(picture_format)}
        NAME {ccccca* b(name_attr1) b(name_attr2) b(shortcut) 
            len1 len2 rest ; 
            set b(text_of_name) [string range $rest  0 [expr $len1-1]] ;
            set b(parsed_expr_for_name) [string range $rest $len1 [expr $len1+len2-1]]
        }
        DIMENSION {ssss b(first_defined_row) b(last_defined_row+1) \
                       b(first_defined_col) b(last_defined_col+1)}
        COLUMN_DEFAULT {ssa* b(column_default_min) b(column_default_max) \
                            b(column_default_cell_attr)}
        ROW {ssssscsa3  b(row_number) b(first_col_in_row) \
                 b(last_col_in_row+1) b(row_height) b(mirco_crap) \
                 b(has_default_attr)  b(rel_file_offset_to_cell) \
                 b(default_cell_attr)}
        BLANK {ssa3 b(row) b(column) b(cell_attr)}
        INTEGER {ssa3s b(row) b(column) b(cell_attr) b(value) }
        NUMBER { ssa3a8 b(row) b(column) b(cell_attr) b(value) }
        LABEL { ssa3ca* b(row) b(column) b(cell_attr) b(len) b(value) }
        BOOLERR {ssa3cc b(row) b(column) b(cell_attr) b(value) b(is_error)}
        FORMULA {ssa3a8ca* b(row) b(column) b(cell_attr) b(value) \
                     b(needs_to_be_calculated) b(len) b(parsed_expression) }
        ARRAY {sscccca* b(row) b(column) b(first_col_of_array) \
                   b(last_col_of_array) b(needs_to_be_calculated) b(len) \
                   b(parsed_expression) }
        CONTINUE {a* b(parsed_expression)}
        STRING {ca* b(len) b(string)}
        TABLE { ssccccss b(first_row_of_table) b(last_row_of_table) \
                    b(first_col_of_table) b(last_col_of_table) \
                    b(needs_to_be_calculated) b(is_row_input_table) \
                    b(row_of_input_cell) b(col_of_input_cell) }
        TABLE2 { ssccccssss b(first_row_of_table) \
                     b(last_row_of_table) b(first_col_of_table) \
                     b(last_col_of_table) b(needs_to_be_calculated) \
                     b(zero) b(row_of_row_input_cell) \
                     b(col_of_row_input_cell)  b(col_of_col_input_cell) }
        PROTECT { s b(is_protected) }
        WINDOW_PROTECT { s b(is_protected) }
        PASSWORD {a* b(encrypted_password) }
        NOTE {sssa* b(row) b(col) b(len) b(note)}
        WINDOW1 {ssssc b(horiz_position) b(vert_position)  \
                     b(width) b(height) b(hidden) }
        WINDOW2 { cccccssci b(display_formulas) \
                      b(display_gridlines) b(display_row_and_col_headers) \
                      b(frozen_panes) b(display_zero_values)  \
                      b(top_visible_row) b(leftmost_visible_col) \
                      b(gridline_in_foreground_color) b(gridline_color) }
        PANE { ssssc b(horiz_pos_of_split) b(vert_pos_of_split) \
                   b(top_visible_row) b(leftmost_visible_col) \
                   b(pane_number) }
        SELECTION { cssssa* b(pane_number) b(row_of_active_cell) \
                        b(col_of_active_cell) b(reference_number) \
                        b(number_of_references) b(array_of_references)}
        BOUNDSHEET { isca* b(offset) b(grbit) b(len) b(sheetname) }
        XF { ssssssss b(font_index) b(format_index) b(style) b(align) b(icv) b(fill) \
                 b(border1) b(border2) }
        EOT {}
    }  
    
    variable selection_referenceRec \
        { sscc b(first_row) b(last_row) b(first_col) \
              b(last_col) }
        
    
    variable  workbookGlobals \
        { FILEPASS CALCCOUNT CALCMODE PRECISION REFMODE
            DELTA ITERATION 1904 BACKUP PRINT_ROW_HEADER PRINT_GRIDLINES
            HORIZONTAL_PAGE_BREAKS VERTICAL_PAGE_BREAKS DEFAULT_ROW_HEIGHT
            FONT FONT2 HEADER FOOTER LEFT_MARGIN RIGHT_MARGIN TOP_MARGIN
            BOTTOM_MARGIN COLWIDTH EXTERNCOUNT EXTERNSHEET EXTERNNAME
            FORMAT NAME COLUMN_DEFAULT PROTECT
            WINDOW_PROTECT PASSWORD NOTE WINDOW1 PANE SELECTION }
    variable worksheetGlobals \
        { INDEX DIMENSION WINDOW2 }

    set version_loaded 2
}

proc Biff::Biff3 {} {
    variable version_loaded
    if { $version_loaded == 3 } {
        return
    }
    variable typeNames
    array set typeNames {
        521 BOF3
    }
    variable binRec
    array set binRec {
        BOF3 {sss b(version_number) b(document_type) b(unused)}
    }
    Biff2
    set version_loaded 3

}

proc Biff::Biff4 {} {
    variable version_loaded
    if { $version_loaded == 4 } {
        return
    }
    variable typeNames
    array set typeNames {
        1033 BOF4
    }
    variable binRec
    array set binRec {
        BOF4 {sss b(version_number) b(document_type) b(unused)}
    }
    Biff3
    set version_loaded 4
}

proc Biff::Biff5 {} {
    variable version_loaded
    if { $version_loaded == 5 } {
        return
    }
    variable typeNames
    array set typeNames {
        2057 BOF5
    }
    array set binRec {
        BOF5 { ssss b(version) b(document_type) b(build_identifier) \
                   b(build_year) }
    }
    Biff4
    set version_loaded 5
}

proc Biff::Biff7 {} {
    variable version_loaded
    if  { $version_loaded == 7 } {
        return
    }
    Biff5
    variable binRec
    array set binRec {
        FONT { sssssccccca* b(font_height) b(font_attr) \
                   b(index_to_color_palette) \
                   b(bold_style) b(super_sub_script) b(underline) \
                   b(font_family) b(char_set) b(unused) b(len) b(font_name) }
    }
    set version_loaded 7
}

proc Biff::Biff8 {} {
    variable version_loaded
    if { $version_loaded == 8 } {
        return
    }
    Biff7
    variable binRec
    array set binRec {
        BOF5 { ssssii b(version_number) b(document_type) b(build_identifier) \
                   b(build_year) b(file_history_flags) b(lowest_excel_ver) 
        }
    }
    set version_loaded 8
}

# Biff::read_formula_value reads b(value) if type == FORMULA
# and returns the type and the value (if type != String)
# types are: Float, Error, Boolean and String

proc Biff::read_formula_value { value } {
    binary scan $value ca5s $value type dummy ffff
    if { $ffff } {
        switch $type {
            0 { return [String ""] }
            1 { 
                binary scan $value a2ca* dummy1 bool dummy2
                return [Boolean $bool]
            }
            2 {
                binary scan $value a2ca* dummy1 err dummy2
                return [Error $err]
            }
            default {
                Debug::printMessage error \
                    "Biff::get_formual_value failed. Cannot parse $value (-> type==$type)"
                return ""
            }
        }
    } else {
        binary scan $value f value float
        return [Float $float]
    }
}

# converted a value to a formula value
# the 8 bytes hide a type description 

# types are: Float, String, Boolean and Error
proc Biff::get_formula_value { type value } {
    switch $type { 
        Float {
            return [binary format f $value]
        }
        String {
            return [binary format c8 {0 0 0 0 0 0 255 255}]
        }
        Boolean {
            return [binary format c8 [list 1 0 $value 0 0 0 255 255]]
        }
        Error {
            return [binary format c8 [list 2 0 $value 0 0 0 255 255]]
        }
        default {
            Debug::printMessage error "Biff::get_formula_value: don't know how to convert $type"
        }
    }
}
    

proc Biff::open_biff { filename {mode r} } {
    set fd [open $filename $mode]
    fconfigure $fd -translation binary
    return $fd
}

proc Biff::get_unsigned_int16 { int } {
    return [expr $int & 0xFFFF]
}

proc Biff::get_unsigned_int32 { int } {
    return [expr $int & 0xFFFFFFFF]
}

proc Biff::get_unsigned_int8 { int } {
    return [expr $int & 0xFF]
}


if { $Biff::bigendian } {
    proc Biff::float_to_386_float { float } {
        binary scan [binary format d $float] c8 c_list
        #puts c_list=$c_list
        set swap_list {}
        foreach c $c_list {
            set swap_list [linsert $swap_list 0 $c]
        }
        return [binary format c8 $swap_list]
    }
} else {
    proc Biff::float_to_386_float { float } {
        return [binary format d $float]
    }
}
    

proc Biff::get_meaning { varname value } {
    set nameArray ${varname}Name
    variable $nameArray
    set meaning ${nameArray}($value)
    if { [info exists $meaning] } {
        return [set $meaning]
    } else {
        variable [set meaning ${varname}Flags]
        if { [info exists $meaning] } {
            #puts m=$meaning
            set res ""
            foreach key [array names $meaning] {
                #puts key=$key,value=$value
                if { $key & $value } {
                    if { $res != "" } {
                        append res ,
                    }
                    append res [set ${meaning}($key)]
                }
                return "$res ($value)"
            }
        }
        return $value
    }
}


proc Biff::read_record { fd } {
    if { ![eof $fd] } {
        set id [read $fd 4]
        binary scan $id ss type size
        if { [info exists size] } {
            set size [get_unsigned_int16 $size]
            #puts "read: size=$size"
            set data [read $fd $size]
            #puts "read_record: [get_unsigned_int16 $type] $data"
            return [list [get_unsigned_int16 $type] $data]
        }
    }
    return {}
}

proc Biff::rec_header { type data } {
    set len [string length $data]

    if { $len > 2048 } {
        Debug::printMessage error "BIFF record too large: $len (max 2048)"
    }

    return [binary format ss $type $len]
}

proc Biff::write_record { fd type data } {
    puts -nonewline $fd "[rec_header $data]$data"
}


proc Biff::search_BOF { fd } {
    while 1 {
        if { [read $fd 1] == "\x09" } {
            seek $fd -1 current
            return
        } elseif { [eof $fd] } {
            return
        }
    }
}

proc Biff::print_records { file } {
    variable typeNames
    variable binRec
    set fd [open_biff $file]
    #search_BOT $fd
    while { 1 } {
        foreach {type data} [read_record $fd] {}
        if { [eof $fd] } {
            break
        }
        if { ![info exists typeNames($type)] } {
            puts "unknown type: $type"
        } else {
            set typename $typeNames($type)
            puts -nonewline "$typename: "
            set parseInfo $binRec($typename)
            if { $parseInfo == {} } {
                if { $data == "" } {
                    puts "(record doesn't contains additional data)"
                } else {
                    puts "don't know how to parse the data"
                }
            } else {
                eval binary scan {$data} $binRec($typename)
                set separator ""
                foreach index [array names b] {
                    puts -nonewline \
                        "$separator $index=[get_meaning $index $b($index)]"
                    set separator ,
                    unset b($index)
                }
                puts ""
            }
        }
    }
    close $fd
}

proc Biff::get_xf { style } {
    if { $stype == "style" } {
        set b(style) 65525
    } else {
        set b(style) 1
    }
    set b(ifnt) 0
    set b(ifmt) 0
    set b(align) 0
}
                   

proc Biff::new_table { tablename { biffversion BIFF8 } } {
    upvar $tablename mytable
    set mytable(BIFF) $biffversion
    set mytable(BOF) {
        document_type 16
        version_number 2
    }
    set fontrec {
        font_height 200
        font_attr 0
        index_to_color_palette 0
        bold_style 0
        super_sub_script 0
        underline 0
        char_set 0
        unused 0
        font_family 0
        font_name Arial
        len 5
    }
    set mytable(DIMENSION) {{
        first_defined_row 0
        last_defined_row+1 100
        first_defined_col 0
        last_defined_col+1 100
    }}
    set mytable(EOT)  {}
    switch $biffversion {
        BIFF8 {
            set mytable(FONT) {}
            for { set i 0 } { $i <= 5 } { incr i } {
                lappend mytable(FONT) $fontrec
            }
            #set mytable(WINDOW1) { TODO }
            set mytable(XF) {}
            for { set i 0 } { $i <= 15 } { incr i } {
                lappend mytable(XF) \
                    [list \
                         font_index 0 \
                         format_index 0 \
                         style 0 \
                         align 0 \
                         icv 0 \
                         fill 0 \
                         border1 0 \
                         border2 0]
            }
            #set mytable(STYLE) {  } ??
        }
    }
}

proc Biff::new_sheet { table sheet } {
    upvar $sheet mysheet
    upvar $table mytable
    set mysheet(DIMENSION) {{
        first_defined_row 0
        last_defined_row+1 100
        first_defined_col 0
        last_defined_col+1 100
    }}
    set mysheet(data) ""
    if { $mytable(BIFF) >= "BIFF5" } {
        ## set mysheet(INDEX) ??? Biff.cpp
    }
}

proc Biff::get_recordsize { data } {
    return [expr 4+[string length $data]]
}

proc Biff::get_record { what barr } {
    variable binRec
    #puts [array get binRec]
    upvar $barr b
    set translate [lindex $binRec($what) 0]
    foreach var [lrange $binRec($what) 1 end] {
        lappend translate [set $var]
    }
    if { $translate != "" } {
        set data [eval binary format $translate]
    } else {
        set data ""
    }
    variable typeNums
    return "[rec_header $typeNums($what) $data]$data"
}

proc Biff::write_biff { fd what barr } {
    upvar $barr mb
    write_record $fd $typeNum($what) [get_record $what mb]
}

proc Biff::get_record_data { what blist } {
    #puts blist=$blist
    array set b $blist
    return [get_record $what b]
}

proc Biff::globals_data { tablename globals } {
    upvar $tablename mytable
    #variable workbookGlobals
    variable binRec
    variable typeNums

    set result ""
    foreach recordType $globals {
        set type $typeNums($recordType)
        if { [info exists mytable($recordType)] } {
            foreach record $mytable($recordType) {
                append result [get_record_data $recordType $record]
            }
        }
    }
    return $result
}

proc Biff::sheetContents_data { sheet } {
    upvar $sheet mycells
    set result {}
    #puts a=[array names mycells]
    foreach cell [array names mycells *,*] {
        foreach {b(column) b(row)} [split $cell ,] {}
        switch -regexp -- $mycells($cell) {
            {^[0-9]+$} {
                set type INTEGER
                set b(value) $mycells($cell)
            }
            {^[0-9]*\.[0-9]+$} {
                set type NUMBER
                set b(value) [float_to_386_float $mycells($cell)]
            }
            ^$ {
                unset mycells($cell)
                continue
            }
            default {
                # only BIFF2
                set type LABEL
                set b(value) $mycells($cell)
                set b(len) [string length $b(value)]
            }
        }
        set b(cell_attr) \x00\x00\x00
        #puts b(value)=$b(value)
        append result [get_record $type b]
        unset mycells($cell)
    }
    append mycells(data) $result
    return $mycells(data)
}

proc Biff::pack_some_data { sheet } {
    upvar $sheet mysheet
    sheetContents_data mysheet
}

proc Biff::BOF_data { num attr } {
    return [get_record_data BOF$num $attr]
}

proc Biff::EOT_data { tablename } {
    upvar $tablename mytable
    return [get_record_data EOT $mytable(EOT)]
}

proc Biff::BOUNDSHEET_data { sheet_name offset } {
    return [get_record_data BOUNDSHEET \
                [list \
                     grbit 0 \
                     offset $offset \
                     len [string length $sheet_name] \
                     sheetname $sheet_name]]
}
                 
                 
proc Biff::get_biff2_table { tablename sheets } {
    upvar $tablename mytable
    
    Biff2

    if { [llength $sheets] > 1 } {
        Debug::printMessage error \
            "BIFF2 supports only 1 sheet"
    }
    upvar [lindex [lindex $sheets 0] 1] mysheet
    #puts $sheets,[array names mysheet]
    puts [lindex $sheets 1]
    set data {}
    append data [BOF_data "" $mytable(BOF)]
    variable workbookGlobals
    append data [globals_data mytable $workbookGlobals]
    variable worksheetGlobals
    append data [globals_data mytable $worksheetGlobals]
    append data [sheetContents_data mysheet]
    append data [EOT_data mytable]
    
    return $data
}



proc Biff::get_biff8_table {tablename sheets} {
    # this is biff7 ;-)
    upvar $tablename mytable

    Biff8
    
    set data {}
    append data [BOF_data 5 {
        version_number 1280
        document_type 5
        build_identifier 0
        build_year 2001
        file_history_flags 0
        lowest_excel_ver 0
    }]
    variable workbookGlobals
    append data [globals_data mytable $workbookGlobals]
    set sheetdata {}
    set i 0
    set boundsheet_size 0
    variable worksheetGlobals
    foreach sheet $sheets {
        set offset($i) [string length $sheetdata]
        incr boundsheet_size [expr 4+4+2+1+[string length [lindex $sheet 0]]]
        incr i
        append sheetdata [BOF_data 5 {
            version_number 1280
            document_type 16
            build_identifier 2412
            build_year 1993
            file_history_flags 0
            lowest_excel_ver 0
        }]
        upvar [lindex $sheet 1] cells
        append sheetdata [globals_data cells $worksheetGlobals]
        append sheetdata [sheetContents_data cells]
        append sheetdata [EOT_data mytable]
    }
    #append BOUNDSHEET records
    set d_offset [expr [string length $data]+$boundsheet_size+4]
    # +4 ==> EOT_data
    set i 0
    foreach sheet $sheets {
        append data [BOUNDSHEET_data [lindex $sheet 0] \
                         [expr $d_offset+$offset($i)]]
    }
    append data [EOT_data mytable]
    append data $sheetdata
    #append data [EOT_data mytable]
    append data [EOT_data mytable]
}


proc Biff::get_table { tablename sheets } {
    upvar $tablename table 
    foreach sheet $sheets {
        upvar [lindex $sheet 1] [lindex $sheet 1]
    }
    switch -- $table(BIFF) {
        BIFF2 {
            return [get_biff2_table table $sheets]
        }
        BIFF8 {
            return [get_biff8_table table $sheets]
        }
        default {
            Debug::printMessage error \
                "Biff: Version $tablename(BIFF) not supported"
            return {}
        }
    }
}



# proc Biff::write_tableBiff2 { filename tablename sheets {useSheets 1} } {
#     upvar $tablename mytable
#     if { [llength $sheets] > 1 && !$useSheets } {
#         Debug::printMessage error "Biff::write_table: useSheets==0, but more than 1 sheet given."
#     }
#     variable writeOrder 
#     variable binRec
#     variable typeNums
#     set fd [open_biff $filename w]
# #    puts [array names typeNums]
#             foreach sheet $sheets {
#                 incr namelensum [string length \
#                                      [set sheetname($i) [lindex $sheet 0]]]
#                 set cellsname [lindex $sheet 1]
#                 set sheetsize($i) 0
#                 upvar $cellsname mycells
#                 foreach cell [array names mycells] {
#                     foreach {b(row) b(column)} [split $cell ,] {}
#                     set b(value) [float_to_386_float $mycells($cell)]
#                     puts b(value)=$b(value)
#                     set b(cell_attr) \x00\x00\x00
#                     set translate [lindex $binRec(NUMBER) 0]
#                     foreach var [lrange $binRec(NUMBER) 1 end] {
#                         lappend translate [set $var]
#                     }
#                     set data [eval binary format $translate]
#                     lappend cellrecs($i) [list NUMBER $data]
#                     incr sheetsize($i) [get_recordsize $data]
#                     unset b
#                 }
#                 incr i
#             }
#             if { $useSheets } {
#                 write_biff $fd EOT b
#                 # write BOUNDSHEETS records
#                 set reloffset 0
#                 set tell [tell $fd]
#                 for { set j 0 } { $j < $i } { incr j } {
#                     set b(version_nummer) 2 doc
#                     set cellrecs($j) \
#                         [concat \
#                              [list BOT [get_record $
                                      
#                     lappend cellrecs($j) [list EOT \
#                                               [binary format \
#                                                    [lindex $binRec(EOT) 0]]]
#                     set translate [lindex $binRec(BOUNDSHEET) 0]

#                     set b(grbit) 0
#                     set b(offset) [expr $tell+\
#                                        11*$i+$namelensum+$reloffset]
#                     incr reloffset [expr $sheetsize($j)+4]
#                     set b(len) [string length $sheetname($j)]
#                     set b(sheetname) $sheetname($j)
#                     foreach var [lrange $binRec(BOUNDSHEET) 1 end] {
#                         lappend translate [set $var]
#                     }
#                     set data [eval binary format $translate]
#                     write_record $fd $typeNums(BOUNDSHEET) $data
#                 }
#             }
#             for { set j 0 } { $j < $i } { incr j } {
#                 foreach rec $cellrecs($j) {
#                     write_record $fd $typeNums([lindex $rec 0]) [lindex $rec 1]
#                 }
#             }
#     close $fd
#                      }

proc Biff::write_table {name table sheets} {
    set fd [open $name w]
    puts -nonewline $fd [uplevel 1 "Biff::get_table $table [list $sheets]"]
    close $fd
}

# test:
#
proc Biff::test {} {
    global argv
    if { $argv == "" } {
        Biff::new_table tab BIFF2
        Biff::new_sheet tab cells
        #Biff::new_sheet table cells2
        set cells(2,3) 2.3
        set cells(1,0) 10
        set cells(0,1) "Hallo"
        #set cells2(1,1) 3.2
        #Biff::write_table "biff8.xls" table {{MyTabelle1 cells} {MyTabelle2 cells2} }
        Biff::write_table "biff2a.xls" tab {{MyTab cells}}
        Biff::print_records biff2a.xls
    } else {
        puts here
        Biff::Biff2
        puts  [array names Biff::typeNames]
        Biff::print_records $argv
    }
}

proc Biff::setDimension { tab x1 y1 x2 y2 } {
    upvar $tab mytab
    set mytab(DIMENSION) \
        [list \
             [list \
                  first_defined_row $y1 \
                  first_defined_col $x1 \
                  last_defined_row+1 [expr $y2+1] \
                  last_defined_col+1 [expr $x2+1]]]
}
              
if { [file tail $argv0] == "biff.tcl" } {
    Biff::test
    exit 
}
package provide Biff 0.1

